# -*- coding: latin-1 -*-

from pyshacl.validate import validate, Validator

# version compliant with https://www.python.org/dev/peps/pep-0440/
__version__ = '0.11.4'

__all__ = ['validate', 'Validator', '__version__']
